// Function.cs
using System;


namespace TtLisp {
  public abstract class Function : Atom {
    // -- instance -------------------------------------------------------
    private List               lambdaList;
    private SymbolicExpression body;

    protected Function( List lambdaList, SymbolicExpression body ) {
      this.lambdaList = lambdaList;
      this.body = body;
    }

    protected List               LambdaList { get { return this.lambdaList; } }
    protected SymbolicExpression Body       { get { return this.body; } }

    public abstract SymbolicExpression Call( Env env, List args );

    protected SymbolicExpression Yield( Env env, List evaledArgs ) {
      if ( this.LambdaList.Size != evaledArgs.Size ) {
        throw new InvalidArgumentCountException();
      }
      env.PushStack();
      try {
        List ll = this.LambdaList;
        List arg = evaledArgs;
        while ( ll.IsNotNil ) {
          env.SymbolTable.LocalIntern( new Symbol( ((ISymbol)ll.Car).Name, arg.Car ) );
          ll = (List)ll.Cdr;
          arg = (List)arg.Cdr;
        }
        return env.Eval( this.Body );
      }
      catch ( InvalidCastException ) {
        throw new InvalidLambdaListException();
      }
      finally {
        env.PopStack();
      }
    }
  }

  public class Lambda : Function {
    // -- instance -------------------------------------------------------
    public Lambda( List lambdaList, SymbolicExpression body ) : base( lambdaList, body ) {}

    public override SymbolicExpression Call( Env env, List args ) {
      return this.Yield( env, this.EvalEachOfList( env, args ) );
    }

    public List EvalEachOfList( Env env, List list ) {
      if ( list.IsNil ) { return list; }
      return new List( env.Eval( list.Car ), list.Cdr is List ? EvalEachOfList( env, (List)list.Cdr ) : list.Cdr );
    }
  }

  public abstract class SpecialForm : Function {
    // -- instance -------------------------------------------------------
    public SpecialForm( List lambdaList, SymbolicExpression body ) : base( lambdaList, body ) {}
  }

  public class Macro : Function {
    // -- instance -------------------------------------------------------
    public Macro( List lambdaList, SymbolicExpression body ) : base( lambdaList, body ) {}

    public override SymbolicExpression Call( Env env, List args ) {
      return env.Eval( this.Yield( env, args ) );
    }
  }

}
