// List.cs
using System;
using System.Reflection;

namespace TtLisp {
  public class List : SymbolicExpression, ISymbol {
    // -- static ---------------------------------------------------------
    public static new readonly List NIL;
    static List() {
      List nil = new List();
      typeof( List ).InvokeMember(
        "car",
        BindingFlags.SetField | BindingFlags.NonPublic | BindingFlags.Instance,
        null, nil, new object[]{ nil } );
      typeof( List ).InvokeMember(
        "cdr",
        BindingFlags.SetField | BindingFlags.NonPublic | BindingFlags.Instance,
        null, nil, new object[]{ nil } );
      List.NIL = nil;
    }

    // -- instance -------------------------------------------------------
    private readonly SymbolicExpression car;
    private readonly SymbolicExpression cdr;

    public List() : this( List.NIL, List.NIL ) {}
    public List( SymbolicExpression car ) : this( car, List.NIL ) {}
    public List( SymbolicExpression car, SymbolicExpression cdr ) {
      this.car = car;
      this.cdr = cdr;
    }

    public override bool IsNil {
      get { return  this.car == List.NIL && this.cdr == List.NIL; }
    }

    public SymbolicExpression Car {
      get { return this.car; }
    }

    public SymbolicExpression Cdr {
      get { return this.cdr; }
    }

    public SymbolicExpression this[int index] {
      get {
        List cur = this;
        for ( int i = 0; i < index; ++i ) {
          cur = (List)cur.Cdr;
        }
        return cur.Car;
      }
    }
      
    
    public int Size {
      get {
        if ( this.IsNil ) { return 0; }
        if ( !(this.Cdr is List) ) { return 1; }
        return ((List)this.Cdr).Size + 1;
      }
    }

    public string Name {
      get {
        if ( this.IsNotNil ) {
          throw new InvalidArgumentTypeException();
        }
        return "nil";
      }
    }

    public SymbolicExpression Value {
      get {
        if ( this.IsNotNil ) {
          throw new InvalidArgumentTypeException();
        }
        return this;
      }
    }

    public bool IsBound {
      get {
        if ( this.IsNotNil ) {
          throw new InvalidArgumentTypeException();
        }
        return true;
      }
    }
    
    public void Bind( SymbolicExpression s ) {
      if ( this.IsNotNil ) {
        throw new InvalidArgumentTypeException();
      }
      throw new InvalidArgumentTypeException();
    }

    public void Unbind() {
      if ( this.IsNotNil ) {
        throw new InvalidArgumentTypeException();
      }
      throw new InvalidArgumentTypeException();
    }

  }
}
