// Symbol.cs
using System;

namespace TtLisp {
  public interface ISymbol {
    string Name { get; }
    SymbolicExpression Value { get; }
    bool IsBound { get; }
    void Bind( SymbolicExpression s );
    void Unbind();
  }

  public class Symbol : Atom, ISymbol {
    // -- static ---------------------------------------------------------
    public static new readonly ISymbol NIL = List.NIL;

    // -- instance -------------------------------------------------------
    private string             name;
    private SymbolicExpression val;

    public Symbol( string name ) : this( name, null ) {}
    public Symbol( string name, SymbolicExpression val ) {
      this.name = name;
      this.val = val;
    }

    public override bool IsNil { get { return this.name == "nil"; } }

    public string Name { get { return this.name; } }

    public bool IsBound { get { return this.val != null; } }
    public SymbolicExpression Value { get { return this.val; } }
    
    public virtual void Bind( SymbolicExpression s ) {
      if ( this.IsNil ) {
        throw new InvalidArgumentTypeException();
      }
      this.val = s;
    }

    public virtual void Unbind() {
      if ( this.IsNil ) {
        throw new InvalidArgumentTypeException();
      }
      this.val = null;
    }

  }
}
