// SymbolTable.cs
using System;
using System.Collections;

namespace TtLisp {
  public class SymbolTable : IEnumerable {
    private Hashtable self;
    private SymbolTable top;

    public SymbolTable() {
      this.self = new Hashtable();
      this.top = this;
    }
    public SymbolTable( SymbolTable other ) : this() {
      this.self = (Hashtable)other.self.Clone();
      this.top = other.top;
    }

    public virtual int Count { get { return self.Count; } }
    public virtual bool IsSynchronized { get { return self.IsSynchronized; } }

    public virtual ISymbol this[ISymbol key] { get { return this[key.Name]; } }
    public virtual ISymbol this[string key] {
      get {
        ISymbol sym;
        if ( self.ContainsKey( key ) ) {
          sym = (ISymbol)self[key];
        }
        else if ( this.top.ContainsKey( key ) ) {
          sym = (ISymbol)this.top[key];
        }
        else {
          throw new SymbolUnboundException();
        }
        if ( !sym.IsBound ) {
          throw new SymbolUnboundException();
        }
        return sym;
      }
    }
    public virtual ICollection Keys { get { return self.Keys; } }
    public virtual object SyncRoot { get { return self.SyncRoot; } }
    public virtual ICollection Values { get { return self.Values; } }

    public virtual void Intern( ISymbol value ) {
      ( this.ContainsKey( value.Name ) ? self : this.top.self ).Add( value.Name, value );
    }
    public virtual void LocalIntern( ISymbol value ) {
      self.Add( value.Name, value );
    }
    public virtual bool ContainsKey( string key ) {
      return self.ContainsKey( key );
    }
    public virtual bool ContainsValue( ISymbol value ) {
      return self.ContainsValue( value );
    }
    public virtual IEnumerator GetEnumerator() {
      return self.GetEnumerator();
    }
    public virtual void Unintern( string key ) {
      ( this.ContainsKey( key ) ? self : this.top.self ).Remove( key );
    }
  }
}
